# -*- coding: utf-8 -*-
#
# Copyright (c) 2021-2022, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the BSD 3-Clause License:
# http://opensource.org/licenses/BSD-3-Clause
#


from __future__ import annotations


import base64
import collections
import pathlib
import queue
import re
import subprocess
import time
try:
    import jupyter_client
except ImportError:
    jupyter_client = None
from .. import util
from .. import message
from ..code_collections import Session
from ..progress import Progress


_ansi_color_escape_code_re = re.compile('\x1b.*?m')
_version_number_re = re.compile(r'(?P<major>\d+)(?:\.(?P<minor>\d+)(?:\.(?P<patch>\d+))?)?')

kernel_name_aliases: dict[str, str] = {}
kernel_name_collisions: dict[str, set[str]] = collections.defaultdict(set)
if jupyter_client is not None:
    for k, v in jupyter_client.kernelspec.KernelSpecManager().get_all_specs().items():
        for alias in [k.lower(), v['spec']['display_name'].lower(), v['spec']['language'].lower()]:
            if alias in kernel_name_aliases:
                kernel_name_collisions[alias].add(k)
            else:
                kernel_name_aliases[alias] = k
    for alias in kernel_name_collisions:
        kernel_name_collisions[alias].add(kernel_name_aliases[alias])
        if len(kernel_name_collisions[alias]) > 1:
            name_to_version_map = {}
            for k in kernel_name_collisions[alias]:
                match = _version_number_re.search(k)
                if match is None:
                    name_to_version_map[k] = None
                else:
                    major = int(match.group('major'))
                    minor = int(match.group('minor')) if match.group('minor') is not None else 0
                    patch = int(match.group('patch')) if match.group('patch') is not None else 0
                    name_to_version_map[k] = (major, minor, patch)
            if not all(v is not None for v in name_to_version_map.values()):
                del kernel_name_aliases[alias]
            else:
                last_version = None
                for k, v in name_to_version_map.items():
                    if last_version is None or v > last_version:
                        kernel_name_aliases[alias] = k
                        last_version = v
                    elif v == last_version:
                        del kernel_name_aliases[alias]
                        break


mime_type_to_file_extension_map: dict[str, str] = {
    'image/png': 'png',
    'image/jpeg': 'jpg',
    'image/svg+xml': 'svg',
    'application/pdf': 'pdf',
}


_home_path_re_pattern = re.escape(pathlib.Path('~').expanduser().as_posix()).replace('/', r'[\\/]')
_home_path_re = re.compile(_home_path_re_pattern, re.IGNORECASE)


async def exec(session: Session, *, cache_key_path: pathlib.Path, progress: Progress) -> None:
    '''
    Execute code from a session with a Jupyter kernel, attach textual output
    to the code chunks within the session, and save rich output files.
    '''
    # https://jupyter-client.readthedocs.io/en/stable/api/client.html
    # https://jupyter-client.readthedocs.io/en/stable/messaging.html#messages-on-the-iopub-pub-sub-channel

    session.did_exec = True
    progress.session_exec_stage_start(session, stage='run')

    if jupyter_client is None:
        msg = 'Cannot import "jupyter_client" Python module; install it and try again'
        session.errors.append(message.SysConfigError(msg))
        progress.session_exec_stage_end(session, stage='run')
        progress.session_finished(session)
        return

    if jupyter_client.version_info < (6, 1):
        # Require async support
        msg = f'jupyter_client >= 6.1.0 is required; version {jupyter_client.__version__} is installed'
        session.errors.append(message.SysConfigError(msg))
        progress.session_exec_stage_end(session, stage='run')
        progress.session_finished(session)
        return

    kernel_name = kernel_name_aliases.get(session.jupyter_kernel.lower())
    if kernel_name is None:
        if session.jupyter_kernel in kernel_name_collisions:
            msg = (
                f'''Jupyter kernel "{session.jupyter_kernel}" is ambiguous; '''
                f'''could refer to {', '.join(f'"{k}"' for k in kernel_name_collisions[session.jupyter_kernel])}'''
            )
        else:
            msg = f'No Jupyter kernel was found for "{session.jupyter_kernel}"'
        session.errors.append(message.SysConfigError(msg))
        progress.session_exec_stage_end(session, stage='run')
        progress.session_finished(session)
        return

    kernel_manager = jupyter_client.AsyncKernelManager(kernel_name=kernel_name)
    try:
        await kernel_manager.start_kernel(stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    except jupyter_client.kernelspec.NoSuchKernel:
        msg = f'No Jupyter kernel was found for "{session.jupyter_kernel}"'
        session.errors.append(message.SysConfigError(msg))
        progress.session_exec_stage_end(session, stage='run')
        progress.session_finished(session)
        return
    except FileNotFoundError:
        msg = f'Jupyter kernel for "{session.jupyter_kernel}" has been deleted or corrupted'
        session.errors.append(message.SysConfigError(msg))
        progress.session_exec_stage_end(session, stage='run')
        progress.session_finished(session)
        return
    except Exception as e:
        msg = f'Failed to start Jupyter kernel for "{session.jupyter_kernel}":\n{e}'
        session.errors.append(message.SysConfigError(msg))
        progress.session_exec_stage_end(session, stage='run')
        progress.session_finished(session)
        return

    kernel_client = kernel_manager.client()
    kernel_client.start_channels()
    try:
        await kernel_client.wait_for_ready()
    except RuntimeError as e:
        kernel_client.stop_channels()
        await kernel_manager.shutdown_kernel()
        msg = f'Jupyter kernel timed out during setup:\n{e}'
        session.errors.append(message.RunConfigError(msg))
        progress.session_exec_stage_end(session, stage='run')
        progress.session_finished(session)
        return

    try:
        kernel_has_errors = False
        incomplete_cc_stack = []
        for cc in session.code_chunks:
            if kernel_has_errors:
                break
            if cc.output_index != cc.index:
                # If incomplete code, accumulate until complete
                incomplete_cc_stack.append(cc)
                continue
            if not incomplete_cc_stack:
                progress.session_chunk_start_exec(session, chunk=cc)
                cc_jupyter_id = kernel_client.execute(cc.code_str)
            else:
                incomplete_cc_stack.append(cc)
                progress.session_chunk_start_exec(session, chunk=incomplete_cc_stack[0])
                cc_jupyter_id = kernel_client.execute('\n'.join(icc.code_str for icc in incomplete_cc_stack))
            deadline = time.monotonic() + session.jupyter_timeout
            while True:
                try:
                    kernel_msg = await kernel_client.get_iopub_msg(timeout=max(0, deadline - time.monotonic()))
                except queue.Empty:
                    kernel_msg = (f'Jupyter kernel "{kernel_name}" timed out during execution '
                                  f'(jupyter_timeout = {session.jupyter_timeout} s)')
                    cc.errors.append(message.RunConfigError(kernel_msg))
                    kernel_has_errors = True
                    break
                if kernel_msg['parent_header'].get('msg_id') != cc_jupyter_id:
                    continue
                kernel_msg_type = kernel_msg['msg_type']
                kernel_msg_content = kernel_msg['content']
                if kernel_msg_type == 'status' and kernel_msg_content['execution_state'] == 'idle':
                    break
                if kernel_msg_type in ('display_data', 'execute_result'):
                    # Rich output
                    if cc.rich_output is None:
                        cc.rich_output = []
                    rich_output_files = {}
                    rich_output = {'files': rich_output_files, 'data': kernel_msg_content['data']}
                    for mime_type, data in kernel_msg_content['data'].items():
                        file_extension = mime_type_to_file_extension_map.get(mime_type)
                        if file_extension is None:
                            continue
                        if 'name' not in cc.options:
                            file_name = f'''{kernel_name}-{session.name or ''}-{cc.output_index+1:03d}-{len(cc.rich_output)+1:02d}.{file_extension}'''
                        else:
                            file_name = f'''{cc.options['name']}-{len(cc.rich_output)+1}.{file_extension}'''
                        session.files.append(file_name)
                        ro_path = cache_key_path / file_name
                        if file_extension == 'svg':
                            ro_path.write_text(data, encoding='utf8')
                        else:
                            ro_path.write_bytes(base64.b64decode(data))
                        rich_output_files[mime_type] = ro_path.as_posix()
                    cc.rich_output.append(rich_output)
                    rich_output_text = kernel_msg_content['data'].get('text/plain')
                    if rich_output_text:
                        progress.session_chunk_rich_output_text(session, chunk=cc, output=rich_output_text)
                    if rich_output_files:
                        progress.session_chunk_rich_output_files(session, chunk=cc, files=rich_output_files.values())
                elif kernel_msg_type == 'stream':
                    if kernel_msg_content['name'] == 'stdout':
                        cc.stdout_lines.extend(util.splitlines_lf(kernel_msg_content['text']))
                        progress.session_chunk_stdout(session, chunk=cc, output=kernel_msg_content['text'])
                    elif kernel_msg_content['name'] == 'stderr':
                        cc.stderr_lines.extend(util.splitlines_lf(_home_path_re.sub('~', kernel_msg_content['text'])))
                        progress.session_chunk_stderr(session, chunk=cc, output=kernel_msg_content['text'])
                elif kernel_msg_type == 'error':
                    kernel_msg_text = _ansi_color_escape_code_re.sub('', '\n'.join(kernel_msg_content['traceback']))
                    kernel_msg_text = _home_path_re.sub('~', kernel_msg_text)
                    # This is currently treated as a `StderrRunError` and
                    # stored in `stderr_lines`.  For some kernels, it may
                    # make more sense to use `RunError` or further refine the
                    # error system.
                    cc.stderr_lines.extend(util.splitlines_lf(kernel_msg_text))
                    cc.errors.append(message.StderrRunError(cc.stderr_lines))
                    kernel_has_errors = True
                    progress.session_chunk_stderr(session, chunk=cc, output=kernel_msg_text)
            if not incomplete_cc_stack:
                progress.session_chunk_end_exec(session, chunk=cc)
            else:
                progress.session_chunk_end_exec(session, chunk=incomplete_cc_stack[-1])
                incomplete_cc_stack = []
    finally:
        kernel_client.stop_channels()
        await kernel_manager.shutdown_kernel()

    progress.session_exec_stage_end(session, stage='run')
    progress.session_finished(session)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2021-2022, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the BSD 3-Clause License:
# http://opensource.org/licenses/BSD-3-Clause
#


from __future__ import annotations


import json
import locale
import pathlib
import pkgutil
import platform
import re
import shlex
import shutil
import string
import sys
from typing import Optional

import bespon
from . import err
from . import version
from . import util
from .util import KeyDefaultDict




_string_formatter = string.Formatter()

def _check_template(template: str):
    '''
    Check a template string for validity.  Make sure all replacement fields
    are plain ASCII keywords without any format specifiers or conversion
    flags.
    '''
    for literal_text, field_name, format_spec, conversion in _string_formatter.parse(template):
        if field_name is None:
            continue
        if not isinstance(field_name, str):
            if isinstance(field_name, int):
                raise TypeError(f'Invalid field "{field_name}": need ASCII identifier, not integer')
            raise TypeError
        if not field_name.isidentifier() or not field_name.isascii():
            raise TypeError(f'Invalid field "{field_name}": need ASCII identifier')
        if format_spec:
            raise TypeError(
                f'Invalid field "{field_name}": need plain keyword without format specifier "{format_spec}"'
            )
        if conversion:
            raise TypeError(
                f'Invalid field "{field_name}": need plain keyword without conversion flag "{conversion}"'
            )

def _split_template(template: str, field: str) -> list[str]:
    '''
    Split template string at specified field name, removing field including
    `{}` delimiters.  Only supports templates with plain keywords, no format
    specifiers or conversion flags.
    '''
    split = []
    start = 0
    end = 0
    for literal_text, field_name, format_spec, conversion in _string_formatter.parse(template):
        if format_spec or conversion:
            raise TypeError('Template strings with format specifiers or conversion flags are not supported')
        end = template.find(literal_text, end) + len(literal_text)
        if field_name != field:
            if end == len(template):
                split.append(template[start:end])
                break
            continue
        split.append(template[start:end])
        end = template.find('}', end) + 1
        start = end
        if start == len(template):
            split.append('')
            break
    if len(split) == 1:
        raise ValueError(f'Field "{field}" was not found')
    return split




class Language(object):
    '''
    Process language definition and insert default values.
    '''
    def __init__(self, name: str, definition: dict):
        if not isinstance(name, str):
            raise TypeError
        self.name: str = name
        if name.endswith('_repl'):
            name_root = name.rsplit('_', 1)[0]
        else:
            name_root = name

        try:
            definition_str = json.dumps(definition)
        except TypeError:
            raise err.CodebraidError(f'Invalid language definition for "{name}"')
        self.definition_bytes: bytes = definition_str.encode('utf8')

        default_encoding = locale.getpreferredencoding(False)

        try:
            language = definition.pop('language', name_root)
            if not isinstance(language, str):
                raise TypeError
            self.language: str = language

            raw_executable = definition.pop('executable', None)
            if raw_executable is None:
                if name_root == 'python':
                    # Use `sys.executable` if it is on PATH as either `python`
                    # or `python3`.  Otherwise, Windows can have `python3`,
                    # and Arch Linux uses `python`, so use `python3` if it
                    # exists and otherwise `python`.  Set `executable` to
                    # absolute path when possible to avoid issues like
                    # https://github.com/python/cpython/issues/83086.
                    executable = None
                    if sys.executable:
                        sys_executable_path = pathlib.Path(sys.executable)
                    else:
                        sys_executable_path = None
                    which_python = shutil.which('python')
                    if which_python:
                        which_python_path = pathlib.Path(which_python)
                    else:
                        which_python_path = None
                    if sys_executable_path is not None and which_python_path is not None:
                        if sys_executable_path == which_python_path or sys_executable_path == which_python_path.resolve():
                            executable = which_python_path.as_posix()
                    if not executable:
                        which_python3 = shutil.which('python3')
                        if which_python3:
                            which_python3_path = pathlib.Path(which_python3)
                            if sys_executable_path is not None and which_python3_path is not None:
                                if sys_executable_path == which_python3_path or sys_executable_path == which_python3_path.resolve():
                                    executable = which_python3_path.as_posix()
                            if not executable:
                                if (platform.system() == 'Windows' and
                                        'AppData/Local/Microsoft/WindowsApps' in which_python3_path.as_posix()):
                                    if which_python_path:
                                        executable = which_python_path.as_posix()
                                    else:
                                        executable = 'python'
                                else:
                                    executable = which_python3_path.as_posix()
                        elif which_python_path:
                            executable = which_python_path.as_posix()
                        else:
                            executable = 'python'
                else:
                    executable = name_root
            elif isinstance(raw_executable, str):
                executable = pathlib.Path(raw_executable).expanduser().as_posix()
                if raw_executable.startswith('./') or raw_executable.startswith('.\\'):
                    executable = f'./{executable}'
            else:
                raise TypeError
            self.executable: str = executable

            raw_interpreter_script = definition.pop('run_script', None)
            if raw_interpreter_script is None:
                interpreter_script = None
            else:
                if not isinstance(raw_interpreter_script, str):
                    raise TypeError
                if any(x in raw_interpreter_script for x in ('..', '\\')):
                    # No relative or Windows-style paths
                    raise ValueError
                scripts_root_dir = pathlib.Path(__file__).parent.resolve() / 'languages' / 'scripts'
                interpreter_script = scripts_root_dir / raw_interpreter_script
                if not interpreter_script.is_file():
                    version_info = version.__version_info__
                    version_num = f'{version_info.major}.{version_info.minor}.{version_info.micro}'
                    scripts_root_dir = pathlib.Path(f'~/.codebraid/{version_num}/languages/scripts').expanduser()
                    interpreter_script = scripts_root_dir / raw_interpreter_script
                    if version_info.releaselevel != 'final' or not interpreter_script.is_file():
                        script_bytes = pkgutil.get_data('codebraid', f'languages/scripts/{raw_interpreter_script}')
                        if script_bytes is None:
                            raise err.CodebraidError(
                                f'Missing interpreter script in language definition for "{name}"'
                            )
                        interpreter_script.parent.mkdir(parents=True, exist_ok=True)
                        interpreter_script.write_bytes(script_bytes)
            self.interpreter_script: Optional[pathlib.Path] = interpreter_script

            raw_opts = definition.pop('executable_opts', None)
            if self.interpreter_script and raw_opts is not None:
                raise TypeError
            if raw_opts is None:
                opts = raw_opts
            elif isinstance(raw_opts, str):
                opts = shlex.split(raw_opts)
                if not opts:
                    raise TypeError
            elif isinstance(raw_opts, list) and raw_opts and all(isinstance(x, str) for x in raw_opts):
                opts = raw_opts
            else:
                raise TypeError
            self.executable_opts: Optional[list[str]] = opts

            raw_args = definition.pop('args', None)
            if self.interpreter_script and raw_args is not None:
                raise TypeError
            if raw_args is None:
                args = raw_args
            elif isinstance(raw_args, str):
                args = shlex.split(raw_args)
                if not args:
                    raise TypeError
            elif isinstance(raw_args, list) and raw_args and all(isinstance(x, str) for x in raw_args):
                args = raw_args
            else:
                raise TypeError
            self.args: Optional[list[str]] = args

            extension = definition.pop('extension', None)
            if not isinstance(extension, str):
                raise TypeError
            if extension.startswith('.'):
                raise ValueError
            self.extension: str = extension

            compile_encoding = definition.pop('compile_encoding', default_encoding)
            if compile_encoding is not None and not isinstance(compile_encoding, str):
                raise TypeError
            self.compile_encoding: Optional[str] = compile_encoding
            compile_commands = definition.pop('compile_commands', [])
            if isinstance(compile_commands, str):
                compile_commands = [compile_commands]
            elif not isinstance(compile_commands, list) or not all(isinstance(x, str) for x in compile_commands):
                raise TypeError
            self.compile_commands: list[str] = compile_commands

            pre_run_encoding = definition.pop('pre_run_encoding', default_encoding)
            if pre_run_encoding is not None and not isinstance(pre_run_encoding, str):
                raise TypeError
            self.pre_run_encoding: Optional[str] = pre_run_encoding
            pre_run_commands = definition.pop('pre_run_commands', [])
            if isinstance(pre_run_commands, str):
                pre_run_commands = [pre_run_commands]
            elif not isinstance(pre_run_commands, list) or not all(isinstance(x, str) for x in pre_run_commands):
                raise TypeError
            self.pre_run_commands: list[str] = pre_run_commands

            run_encoding = definition.pop('run_encoding', default_encoding)
            if run_encoding is not None and not isinstance(run_encoding, str):
                raise TypeError
            self.run_encoding: Optional[str] = run_encoding
            if not self.interpreter_script:
                run_command = definition.pop('run_command', '{executable} {executable_opts} {source} {args}')
            else:
                run_command = definition.pop('run_command', '{executable} {run_script} {run_delim_start} {run_delim_hash} {buffering}')
            if not isinstance(run_command, str):
                raise TypeError
            self.run_command: str = run_command

            post_run_encoding = definition.pop('post_run_encoding', default_encoding)
            if post_run_encoding is not None and not isinstance(post_run_encoding, str):
                raise TypeError
            self.post_run_encoding: Optional[str] = post_run_encoding
            post_run_commands = definition.pop('post_run_commands', [])
            if isinstance(post_run_commands, str):
                post_run_commands = [post_run_commands]
            elif not isinstance(post_run_commands, list) or not all(isinstance(x, str) for x in post_run_commands):
                raise TypeError
            self.post_run_commands: list[str] = post_run_commands

            repl = definition.pop('repl', name.endswith('_repl'))
            if not isinstance(repl, bool):
                raise TypeError
            self.repl: bool = repl

            run_template = definition.pop('run_template', '{code}\n')
            if not isinstance(run_template, str):
                raise TypeError
            if not run_template.endswith('\n'):
                raise ValueError('run_template must end with a newline ("\n")')
            _check_template(run_template)
            try:
                raw_run_template_before_code, raw_run_template_after_code = _split_template(run_template, 'code')
            except ValueError:
                raise ValueError('Run template must contain one and only one "{code}" field')
            raw_run_template_before_code_last_nl_index = raw_run_template_before_code.rfind('\n')
            if raw_run_template_before_code_last_nl_index < 0:
                run_template_before_code = ''
                run_template_before_code_last_line = raw_run_template_before_code
            else:
                run_template_before_code = raw_run_template_before_code[:raw_run_template_before_code_last_nl_index+1]
                run_template_before_code_last_line = raw_run_template_before_code[raw_run_template_before_code_last_nl_index+1:]
            raw_run_template_after_code_first_nl_index = raw_run_template_after_code.find('\n')
            if raw_run_template_after_code_first_nl_index < 0:
                # Technically, this duplicates the run_template.endswith('\n')
                # test from earlier.  Check anyway in case the earlier
                # test/restriction is ever relaxed.
                raise ValueError('Run template must end with a newline ("\n")')
            run_template_after_code = raw_run_template_after_code[raw_run_template_after_code_first_nl_index+1:]
            run_template_after_code_first_line = raw_run_template_after_code[:raw_run_template_after_code_first_nl_index]
            if run_template_before_code_last_line.strip(' \t') or run_template_after_code_first_line.strip(' \t'):
                raise ValueError('In run template, "{code}" field must be on a line by itself')
            run_template_code_indent = run_template_before_code_last_line
            self.run_template_before_code: str = run_template_before_code
            self.run_template_after_code: str = run_template_after_code
            self.run_template_before_code_n_lines: int = run_template_before_code.count('\n')
            self.run_template_after_code_n_lines: int = run_template_after_code.count('\n')

            if self.interpreter_script:
                if not self.repl:
                    chunk_wrapper_default = (
                        '{stdout_start_delim}\n'
                        '{stderr_start_delim}\n'
                        '{code}\n'
                        '{stdout_end_delim}\n'
                        '{stderr_end_delim}\n'
                    )
                else:
                    chunk_wrapper_default = (
                        '{stdout_start_delim}\n'
                        '{stderr_start_delim}\n'
                        '{repl_start_delim}\n'
                        '{code}\n'
                        '{repl_end_delim}\n'
                        '{stdout_end_delim}\n'
                        '{stderr_end_delim}\n'
                    )
                chunk_wrapper = definition.pop('chunk_wrapper', chunk_wrapper_default)
            else:
                chunk_wrapper = definition.pop('chunk_wrapper')
            if not isinstance(chunk_wrapper, str):
                raise TypeError
            if not chunk_wrapper.endswith('\n'):
                raise ValueError
            _check_template(chunk_wrapper)
            try:
                raw_chunk_wrapper_before_code, raw_chunk_wrapper_after_code = _split_template(chunk_wrapper, 'code')
            except ValueError:
                raise ValueError('Chunk wrapper template must contain one and only one "{code}" field')
            raw_chunk_wrapper_before_code_last_nl_index = raw_chunk_wrapper_before_code.rfind('\n')
            if raw_chunk_wrapper_before_code_last_nl_index < 0:
                raw_chunk_wrapper_before_code = ''
                raw_chunk_wrapper_before_code_last_line = raw_chunk_wrapper_before_code
            else:
                raw_chunk_wrapper_before_code = raw_chunk_wrapper_before_code[:raw_chunk_wrapper_before_code_last_nl_index+1]
                raw_chunk_wrapper_before_code_last_line = raw_chunk_wrapper_before_code[raw_chunk_wrapper_before_code_last_nl_index+1:]
            raw_chunk_wrapper_after_code_first_nl_index = raw_chunk_wrapper_after_code.find('\n')
            if raw_chunk_wrapper_after_code_first_nl_index < 0:
                # Technically, this duplicates the chunk_wrapper.endswith('\n')
                # test from earlier.  Check anyway in case the earlier
                # test/restriction is ever relaxed.
                raise ValueError('Chunk wrapper template must end with a newline ("\n")')
            raw_chunk_wrapper_after_code = raw_chunk_wrapper_after_code[raw_chunk_wrapper_after_code_first_nl_index+1:]
            raw_chunk_wrapper_after_code_first_line = raw_chunk_wrapper_after_code[:raw_chunk_wrapper_after_code_first_nl_index]
            if raw_chunk_wrapper_before_code_last_line.strip(' \t') or raw_chunk_wrapper_after_code_first_line.strip(' \t'):
                raise ValueError('In chunk wrapper template, "{code}" field must be on a line by itself')
            raw_chunk_wrapper_code_indent = raw_chunk_wrapper_before_code_last_line
            chunk_wrapper_code_indent = run_template_code_indent + raw_chunk_wrapper_code_indent
            self.chunk_wrapper_code_indent: str = chunk_wrapper_code_indent
            raw_chunk_wrapper_before_code_lines = util.splitlines_lf(raw_chunk_wrapper_before_code)
            raw_chunk_wrapper_after_code_lines = util.splitlines_lf(raw_chunk_wrapper_after_code)
            for lines in (raw_chunk_wrapper_before_code_lines, raw_chunk_wrapper_after_code_lines):
                if all(not line or line.startswith((' ', '\t')) for line in lines):
                    raise ValueError(
                        'Chunk wrapper template must not be indented; indentation is inherited from run template'
                    )
            self.chunk_wrapper_before_code: str = ''.join(f'{run_template_code_indent}{line}\n' for line in raw_chunk_wrapper_before_code_lines)
            self.chunk_wrapper_before_code_n_lines: int = len(raw_chunk_wrapper_before_code_lines)
            self.chunk_wrapper_after_code: str = ''.join(f'{run_template_code_indent}{line}\n' for line in raw_chunk_wrapper_after_code_lines)
            self.chunk_wrapper_after_code_n_lines: int = len(raw_chunk_wrapper_after_code_lines)

            raw_inline_expr_fmter = definition.pop('inline_expression_formatter', None)
            if raw_inline_expr_fmter is None:
                inline_expression_formatter = None
                inline_expr_fmter_before_code_n_lines = None
                inline_expr_fmter_n_lines = None
            else:
                if not isinstance(raw_inline_expr_fmter, str):
                    raise TypeError
                if not raw_inline_expr_fmter.endswith('\n'):
                    raise ValueError
                _check_template(raw_inline_expr_fmter)
                try:
                    raw_inline_expr_fmter_before_code, _ = _split_template(raw_inline_expr_fmter, 'code')
                except ValueError:
                    raise ValueError(
                        'Inline expression formatter template must contain one and only one "{code}" field'
                    )
                inline_expr_fmter_before_code_n_lines = raw_inline_expr_fmter_before_code.count('\n')
                raw_inline_expr_fmter_lines = util.splitlines_lf(raw_inline_expr_fmter)
                if all(not line or line.startswith((' ', '\t')) for line in raw_inline_expr_fmter_lines):
                    raise ValueError(
                        'Inline expression formatter template must not be indented; indentation is inherited from other templates'
                    )
                # Inline expression formatter has no indent to consider beyond
                # chunk wrapper indent, since code is only ever a single line.
                inline_expression_formatter = ''.join(f'{chunk_wrapper_code_indent}{line}\n' for line in raw_inline_expr_fmter_lines)
                inline_expr_fmter_n_lines = len(raw_inline_expr_fmter_lines)
            self.inline_expression_formatter: Optional[str] = inline_expression_formatter
            self.inline_expression_formatter_n_lines: Optional[int] = inline_expr_fmter_n_lines
            self.inline_expression_formatter_before_code_n_lines: Optional[int] = inline_expr_fmter_before_code_n_lines

            error_patterns = definition.pop('error_patterns', ['error', 'Error', 'ERROR'])
            if isinstance(error_patterns, str):
                error_patterns = [error_patterns]
            elif not (isinstance(error_patterns, list) and error_patterns and
                      all(isinstance(x, str) and x for x in error_patterns)):
                raise TypeError
            self.error_patterns: list[str] = error_patterns
            warning_patterns = definition.pop('warning_patterns', ['warning', 'Warning', 'WARNING'])
            if isinstance(warning_patterns, str):
                warning_patterns = [warning_patterns]
            elif not (isinstance(warning_patterns, list) and warning_patterns and
                      all(isinstance(x, str) and x for x in warning_patterns)):
                raise TypeError
            self.warning_patterns: list[str] = warning_patterns
            line_number_raw_patterns = definition.pop('line_number_patterns', [':{number}', 'line {number}'])
            if line_number_raw_patterns is None:
                pass
            elif isinstance(line_number_raw_patterns, str):
                line_number_raw_patterns = [line_number_raw_patterns]
            elif not (isinstance(line_number_raw_patterns, list) and line_number_raw_patterns and
                      all(isinstance(x, str) and x for x in line_number_raw_patterns)):
                raise TypeError
            self.line_number_raw_patterns: Optional[list[str]] = line_number_raw_patterns
            line_number_regex = definition.pop('line_number_regex', None)
            if line_number_regex is not None and not isinstance(line_number_regex, str):
                raise TypeError
            self.line_number_regex: Optional[str] = line_number_regex
            if line_number_raw_patterns is None and line_number_regex is None:
                raise TypeError
        except KeyError as e:
            raise err.CodebraidError(f'Missing key(s) in language definition for "{name}":\n{e}')
        except (TypeError, ValueError) as e:
            raise err.CodebraidError(f'Invalid data type or value in language definition for "{name}":\n{e}')
        if definition:
            unknown_keys = ', '.join(f'"{k}"' for k in definition)
            raise err.CodebraidError(f'Unknown key(s) in language definition for "{name}": {unknown_keys}')

        self.exec_stages = {}
        if self.compile_commands:
            self.exec_stages['compile'] = self.compile_commands
        if self.pre_run_commands:
            self.exec_stages['pre_run'] = self.pre_run_commands
        if self.run_command:
            self.exec_stages['run'] = self.run_command
        if self.post_run_commands:
            self.exec_stages['post_run'] = self.post_run_commands

        if line_number_raw_patterns:
            line_number_patterns = []
            for lnp in line_number_raw_patterns:
                try:
                    lnp_split = _split_template(lnp, 'number')
                except ValueError as e:
                    raise err.CodebraidError(f'Invalid line number pattern in language definition for "{name}":\n{e}')
                line_number_patterns.append(r'(\d+)'.join(re.escape(x) for x in lnp_split))
            line_number_pattern = '|'.join(line_number_patterns)
            try:
                line_number_pattern_re =  re.compile(line_number_pattern)
            except Exception as e:
                raise err.CodebraidError(f'Invalid line number pattern in language definition for "{name}":\n{e}')
        else:
            line_number_pattern = None
            line_number_pattern_re = None
        self.line_number_pattern: Optional[str] = line_number_pattern
        self.line_number_pattern_re: Optional[re.Pattern] = line_number_pattern_re
        if self.line_number_regex is None:
            line_number_regex_re = None
        else:
            try:
                line_number_regex_re = re.compile(self.line_number_regex, re.MULTILINE)
            except Exception as e:
                raise err.CodebraidError(f'Invalid line number regex in language definition for "{name}":\n{e}')
        self.line_number_regex_re: Optional[re.Pattern] = line_number_regex_re




templates_root = 'languages'
_raw_index = pkgutil.get_data('codebraid', f'{templates_root}/index.bespon')
if _raw_index is None:
    raise err.CodebraidError(f'Failed to find "codebraid/{templates_root}/index.bespon"')
index = bespon.loads(_raw_index)
del _raw_index


def _load_language(lang_name: str) -> Optional[Language]:
    '''
    Load and return language definition from if it exists, else return None.
    '''
    try:
        lang_def_filename = index[lang_name]
    except KeyError:
        return None
    lang_def_bytes = pkgutil.get_data('codebraid', f'{templates_root}/{lang_def_filename}')
    if lang_def_bytes is None:
        return None
    try:
        lang_def = bespon.loads(lang_def_bytes)
    except Exception as e:
        raise err.CodebraidError(
            f'Failed to load language definition for "{lang_name}" (invalid or corrupted definition?):\n{e}'
        )
    return Language(lang_name, lang_def[lang_name])
languages = KeyDefaultDict(_load_language)
del _load_language



# -*- coding: utf-8 -*-
#
# Copyright (c) 2021-2022, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the BSD 3-Clause License:
# http://opensource.org/licenses/BSD-3-Clause
#


from __future__ import annotations


import hashlib
import json
import pathlib
import shlex
from typing import Dict, List, NamedTuple, Optional, Union
from . import language
from . import message
from .code_chunks import CodeChunk, CodeKey
from .codebraid_defaults import CodebraidDefaults




class CodeLineOrigin(NamedTuple):
    '''
    Track origin of line of code with code chuck and (user code) line number.
    This is mainly for synchronizing error and warning messages for code that
    is executed.
    '''
    chunk: Optional[CodeChunk]
    line_number: int




class CodeCollection(object):
    '''
    Base class for sessions and sources.  Never instantiated.
    '''
    def __init__(self, code_key: CodeKey, *, codebraid_defaults: CodebraidDefaults):
        if type(self) is CodeCollection:
            raise NotImplementedError
        self.key: CodeKey = code_key
        self.lang: Optional[str] = code_key.lang
        self.name: Optional[str] = code_key.name
        self.origin_name: Optional[str] = code_key.origin_name
        self.codebraid_defaults: CodebraidDefaults = codebraid_defaults

        self.code_chunks: List[CodeChunk] = []
        self._code_start_line_number: int = 1
        self.code_chunk_origins: set[str] = set()
        self.code_chunk_placeholder_langs = set()

        self.status: message.CodeStatus = message.CodeStatus()
        self.errors: message.ErrorMessageList = message.ErrorMessageList(status=self.status)
        self.warnings: message.WarningMessageList = message.WarningMessageList(status=self.status)
        self.files: List[str] = []

    type: str = 'code'
    _optional_attrs = set(['live_output'])




class Source(CodeCollection):
    '''
    An ordered collection of code chunks that is typically displayed but never
    executed.  May be exported as a single file of source code.
    '''
    def __init__(self, code_key: CodeKey, *, codebraid_defaults: CodebraidDefaults):
        super().__init__(code_key, codebraid_defaults=codebraid_defaults)

    type: str = 'source'


    def append(self, code_chunk: CodeChunk):
        '''
        Append a code chunk to internal code chunk list.  Check code chunk
        options for validity and update chunk summary data.
        '''
        code_chunk.code_collection = self
        code_chunk.index = len(self.code_chunks)
        code_chunk.errors.register_status(self.status)
        code_chunk.warnings.register_status(self.status)
        first_chunk_options = code_chunk.options['first_chunk_options']
        if first_chunk_options:
            invalid_options = ', '.join(f'"{k}"' for k in first_chunk_options)
            first_chunk_options.clear()
            msg = f'Some options are only valid for a session, not for a source: {invalid_options}'
            code_chunk.errors.append(message.SourceError(msg))
        self.code_chunks.append(code_chunk)
        self.code_chunk_origins.add(code_chunk.origin_name)
        if 'placeholder_lang' in code_chunk.options:
            self.code_chunk_placeholder_langs.add(code_chunk.options['placeholder_lang'])


    def finalize(self):
        '''
        Perform tasks that must wait until all code chunks are present.
        '''
        for cc in self.code_chunks:
            if not cc.inline:
                cc.finalize_line_numbers(self._code_start_line_number)
                # Only named sources have sequential line numbering between
                # code chunks
                if self.name is not None:
                    self._code_start_line_number += len(cc.code_lines)




class Session(CodeCollection):
    '''
    Code chunks comprising a session.
    '''
    def __init__(self, code_key: CodeKey, *, codebraid_defaults: CodebraidDefaults):
        super().__init__(code_key, codebraid_defaults=codebraid_defaults)

        self.lang_def: Optional[language.Language] = None
        self.executable: Optional[str] = None
        self.executable_opts: Optional[list[str]] = None
        self.args: Optional[list[str]] = None
        self.live_output: Optional[bool] = None
        self.repl: Optional[bool] = None
        self.jupyter_kernel: Optional[str] = None
        self.jupyter_timeout: Optional[int] = None

        self.needs_exec: bool = True
        self.did_exec: bool = False
        self.is_finalized: bool = False

        self.decode_error_count: int = 0
        self.max_tracked_decode_error_count: int = 10

        # Output that isn't directly associated with individual code chunks.
        # This can come from the start or end of the run code template, or
        # from other sources like Python syntax errors that are generated
        # before code ever runs or compilation output from Rust.
        self.compile_lines: list[str] = []
        self.pre_run_output_lines: list[str] = []
        self.template_start_stdout_lines: list[str] = []
        self.template_start_stderr_lines: list[str] = []
        self.template_end_stdout_lines: list[str] = []
        self.template_end_stderr_lines: list[str] = []
        self.other_stdout_lines: list[str] = []
        self.other_stderr_lines: list[str] = []
        self.post_run_output_lines: list[str] = []

        self._run_code: Optional[str] = None
        self.run_delim_start: str = '#Codebraid'
        self.run_delim_start_search_pattern: str= f'{self.run_delim_start}('
        self.run_code_to_origins: Optional[dict[int, CodeLineOrigin]] = None
        self.expected_stdout_start_delim_chunks: Optional[dict[int, int]] = None
        self.expected_stderr_start_delim_chunks: Optional[dict[int, int]] = None
        self.expected_stdout_end_delim_chunks: Optional[dict[int, int]] = None
        self.expected_stderr_end_delim_chunks: Optional[dict[int, int]] = None

    type: str = 'session'

    _default_live_output = False
    _default_jupyter_timeout = 60


    def append(self, code_chunk: CodeChunk):
        '''
        Append a code chunk to internal code chunk list.  Check code chunk
        options for validity and update chunk summary data.
        '''
        code_chunk.code_collection = self
        code_chunk.index = len(self.code_chunks)
        code_chunk.errors.register_status(self.status)
        code_chunk.warnings.register_status(self.status)
        first_chunk_options = code_chunk.options['first_chunk_options']
        if code_chunk.index == 0:
            if any(k.startswith('jupyter') for k in first_chunk_options) or 'jupyter' in self.codebraid_defaults:
                self.repl = False
                self.jupyter_kernel = first_chunk_options.get(
                    'jupyter_kernel',
                    self.codebraid_defaults.get_keypath('jupyter.kernel', self.lang)
                )
                self.jupyter_timeout = first_chunk_options.get(
                    'jupyter_timeout',
                    self.codebraid_defaults.get_keypath('jupyter.timeout', self._default_jupyter_timeout)
                )
            else:
                self.lang_def = language.languages[self.lang]
                if self.lang_def is None:
                    msg = f'Language definition for "{self.lang}" does not exist'
                    self.errors.append(message.SysConfigError(msg))
                else:
                    try:
                        raw_executable = first_chunk_options['executable']
                    except KeyError:
                        executable = self.lang_def.executable
                    else:
                        executable = pathlib.Path(raw_executable).expanduser().as_posix()
                        if raw_executable.startswith('./') or raw_executable.startswith('.\\'):
                            executable = f'./{executable}'
                    self.executable = executable
                    try:
                        raw_executable_opts = first_chunk_options['executable_opts']
                    except KeyError:
                        executable_opts = self.lang_def.executable_opts
                    else:
                        executable_opts = shlex.split(raw_executable_opts)
                    self.executable_opts = executable_opts
                    try:
                        raw_args = first_chunk_options['args']
                    except KeyError:
                        args = self.lang_def.args
                    else:
                        args = shlex.split(raw_args)
                    self.args = args
                    self.repl = self.lang_def.repl
            self.live_output = first_chunk_options.get(
                'live_output',
                self.codebraid_defaults.get('live_output', self._default_live_output)
            )
        elif first_chunk_options:
            invalid_options = ', '.join(f'"{k}"' for k in first_chunk_options)
            first_chunk_options.clear()
            msg = f'Some options are only valid for the first code chunk in a session: {invalid_options}'
            code_chunk.errors.append(message.SourceError(msg))
        self.code_chunks.append(code_chunk)
        self.code_chunk_origins.add(code_chunk.origin_name)
        if 'placeholder_lang' in code_chunk.options:
            self.code_chunk_placeholder_langs.add(code_chunk.options['placeholder_lang'])


    def finalize(self):
        '''
        Perform tasks that must wait until all code chunks are present,
        such as hashing.
        '''
        if self.code_chunks[0].options['outside_main']:
            from_outside_main_switches = 0
        else:
            from_outside_main_switches = 1
        to_outside_main_switches = 0
        incomplete_ccs = []
        last_cc = None
        for cc in self.code_chunks:
            if self.repl and cc.command != 'repl':
                msg = 'Code executed in REPL mode must use the "repl" command'
                cc.errors.append(message.SourceError(msg))
            if cc.is_expr and self.lang_def is not None and self.lang_def.inline_expression_formatter is None:
                msg = f'Inline expressions are not supported for {self.lang_def.name}'
                cc.errors.append(message.SourceError(msg))
            if last_cc is not None and last_cc.options['outside_main'] != cc.options['outside_main']:
                if last_cc.options['outside_main']:
                    from_outside_main_switches += 1
                    if from_outside_main_switches > 1:
                        msg = 'Invalid "outside_main" value; cannot switch back yet again'
                        cc.errors.append(message.SourceError(msg))
                    for icc in incomplete_ccs:
                        # When switching from `outside_main`, all accumulated
                        # output belongs to the last code chunk `outside_main`
                        icc.output_index = last_cc.index
                    incomplete_ccs = []
                else:
                    if not last_cc.options['complete']:
                        msg = 'Final code chunk before switching to "outside_main" must have "complete" value "true"'
                        last_cc.errors.append(message.SourceError(msg))
                    to_outside_main_switches += 1
                    if to_outside_main_switches > 1:
                        msg = 'Invalid "outside_main" value; cannot switch back yet again'
                        cc.errors.append(message.SourceError(msg))
            if cc.options['complete']:
                cc.output_index = cc.index
                if incomplete_ccs:
                    for icc in incomplete_ccs:
                        icc.output_index = cc.index
                    incomplete_ccs = []
            else:
                incomplete_ccs.append(cc)
            last_cc = cc
        if incomplete_ccs:
            if last_cc.options['outside_main']:
                # Last code chunk gets all accumulated output
                for icc in incomplete_ccs:
                    icc.output_index = last_cc.index
            else:
                msg = 'Final code chunk cannot have "complete" value "false"'
                last_cc.errors.append(message.SourceError(msg))
        for cc in self.code_chunks:
            if not cc.inline:
                cc.finalize_line_numbers(self._code_start_line_number)
                self._code_start_line_number += len(cc.code_lines)
        if self.status.prevent_exec:
            # Hashes and line numbers are only needed if code will indeed be
            # executed.  It is impossible to determine these in the case of
            # errors like copy errors which leave code for one or more chunks
            # undefined.
            return

        # The overall hash for a session has the form
        # `<blake2b(session)>_<len(code)>`.  Code is hashed as well as options
        # that affect code execution.  The hashing process needs to use some
        # sort of delimiter between code chunks and between code and its
        # options.  `hasher.digest()` is used for this purpose.  `len(code)`
        # is included in the overall hash as an extra guard against
        # collisions.
        hasher = hashlib.blake2b()
        code_len = 0
        # Hash needs to depend on session to avoid collisions.  Hash needs to
        # depend on options that determine how code is executed.
        if self.executable:
            hashed_options = {
                'session': self.name,
                'executable': self.executable,
                'executable_opts': self.executable_opts,
                'args': self.args,
            }
        elif self.jupyter_kernel:
            hashed_options = {
                'session': self.name,
                'jupyter_kernel': self.jupyter_kernel,
                'jupyter_timeout': self.jupyter_timeout,
            }
        else:
            self.errors.append(message.SourceError(
                'No execution method is specified; missing language or executable or Jupyter kernel'
            ))
            return
        hasher.update(json.dumps(hashed_options).encode('utf8'))
        hasher.update(hasher.digest())
        # Hash needs to depend on the language definition
        if self.lang_def is not None:
            hasher.update(self.lang_def.definition_bytes)
        hasher.update(hasher.digest())
        for cc in self.code_chunks:
            # Hash needs to depend on some code chunk options.  `command`
            # determines some wrapper code.  `inline` affects line count
            # and error sync currently, and might also affect code in the
            # future.  `complete` determines how code is executed as a
            # byproduct of modifying where output appears.
            cc_options = {
                'command': cc.command,
                'inline': cc.inline,
                'complete': cc.options['complete'],
            }
            hasher.update(json.dumps(cc_options).encode('utf8'))
            hasher.update(hasher.digest())
            code_bytes = cc.code_str.encode('utf8')
            hasher.update(code_bytes)
            hasher.update(hasher.digest())
            code_len += len(cc.code_str) + 1  # +1 for omitted trailing newline
        self.hash = f'{hasher.hexdigest()}_{code_len}'
        self.hash_root = self.temp_suffix = hasher.hexdigest()[:16]
        self.run_delim_hash = hasher.hexdigest()[:64]

        self.is_finalized = True


    @property
    def run_code(self) -> str:
        if self._run_code is not None:
            return self._run_code

        if (self.jupyter_kernel is not None or not self.is_finalized or self.status.has_errors):
            raise AttributeError

        # Delim templates for streams
        delim_start = self.run_delim_start
        start_delim_args = f'delim=start, chunk={{chunk}}, output_chunk={{output_chunk}}, hash={self.run_delim_hash},'
        end_delim_args =   f'delim=end, chunk={{chunk}}, output_chunk={{output_chunk}}, hash={self.run_delim_hash},'
        stdout_start_delim_template = f'{delim_start}(output=stdout, {start_delim_args})'
        stdout_end_delim_template =   f'{delim_start}(output=stdout, {end_delim_args})'
        stderr_start_delim_template = f'{delim_start}(output=stderr, {start_delim_args})'
        stderr_end_delim_template =   f'{delim_start}(output=stderr, {end_delim_args})'
        expr_start_delim_template =   f'{delim_start}(output=expr, {start_delim_args})'
        expr_end_delim_template =     f'{delim_start}(output=expr, {end_delim_args})'
        repl_start_delim_template =   f'{delim_start}(output=repl, {start_delim_args})'
        repl_end_delim_template =     f'{delim_start}(output=repl, {end_delim_args})'
        rich_output_start_delim_template = f'{delim_start}(output=rich_output, format={{format}}, {start_delim_args})'
        rich_output_end_delim_template =   f'{delim_start}(output=rich_output, format={{format}}, {end_delim_args})'

        # List of code to execute, plus bookkeeping for tracing errors back to
        # their origin
        run_code_list: list[str] = []
        run_code_line_number: int = 1
        user_code_line_number: int = 1
        self.run_code_to_origins = {}
        self.expected_stdout_start_delim_chunks = {}
        self.expected_stderr_start_delim_chunks = {}
        self.expected_stdout_end_delim_chunks = {}
        self.expected_stderr_end_delim_chunks = {}

        # Assemble code to execute while keeping track of where each line
        # originates
        if not self.code_chunks[0].options['outside_main']:
            run_code_list.append(self.lang_def.run_template_before_code)
            run_code_line_number += self.lang_def.run_template_before_code_n_lines
        last_cc = None
        for cc in self.code_chunks:
            if ((last_cc is not None and last_cc.options['complete']) or
                    (last_cc is not None and last_cc.options['outside_main'] and not cc.options['outside_main'])):
                run_code_list.append(
                    self.lang_def.chunk_wrapper_after_code.format(
                        stdout_end_delim=stdout_end_delim_template.format(chunk=last_cc.index, output_chunk=last_cc.output_index),
                        stderr_end_delim=stderr_end_delim_template.format(chunk=last_cc.index, output_chunk=last_cc.output_index),
                        repl_end_delim=repl_end_delim_template.format(chunk=last_cc.index, output_chunk=last_cc.output_index),
                    )
                )
                run_code_line_number += self.lang_def.chunk_wrapper_after_code_n_lines
                self.expected_stdout_end_delim_chunks[last_cc.index] = 1
                self.expected_stderr_end_delim_chunks[last_cc.index] = 1
            if ((last_cc is None and not cc.options['outside_main']) or
                    (last_cc is not None and last_cc.options['complete']) or
                    (last_cc is not None and last_cc.options['outside_main'] != cc.options['outside_main'])):
                run_code_list.append(
                    self.lang_def.chunk_wrapper_before_code.format(
                        stdout_start_delim=stdout_start_delim_template.format(chunk=cc.index, output_chunk=cc.output_index),
                        stderr_start_delim=stderr_start_delim_template.format(chunk=cc.index, output_chunk=cc.output_index),
                        repl_start_delim=repl_start_delim_template.format(chunk=cc.index, output_chunk=cc.output_index),
                    )
                )
                run_code_line_number += self.lang_def.chunk_wrapper_before_code_n_lines
                self.expected_stdout_start_delim_chunks[cc.index] = 1
                self.expected_stderr_start_delim_chunks[cc.index] = 1
            if cc.inline:
                # Only block code contributes toward line numbers
                if cc.is_expr:
                    expr_start_delim = expr_start_delim_template.format(chunk=cc.index, output_chunk=cc.output_index)
                    expr_end_delim = expr_end_delim_template.format(chunk=cc.index, output_chunk=cc.output_index)
                    expr_code = self.lang_def.inline_expression_formatter.format(
                        expr_start_delim=expr_start_delim,
                        expr_end_delim=expr_end_delim,
                        temp_suffix=self.temp_suffix,
                        code=cc.code_str,
                    )
                    if not self.lang_def.chunk_wrapper_code_indent:
                        run_code_list.append(expr_code)
                    else:
                        run_code_list.append(
                            expr_code.replace('\n', '\n'+self.lang_def.chunk_wrapper_code_indent,
                                              self.lang_def.inline_expression_formatter_n_lines-1)
                        )
                    line_number = run_code_line_number + self.lang_def.inline_expression_formatter_before_code_n_lines
                    self.run_code_to_origins[line_number] = CodeLineOrigin(chunk=cc, line_number=1)
                    run_code_line_number += self.lang_def.inline_expression_formatter_n_lines
                else:
                    run_code_list.append(f'{self.lang_def.chunk_wrapper_code_indent}{cc.code_str}\n')
                    self.run_code_to_origins[run_code_line_number] = CodeLineOrigin(chunk=cc, line_number=1)
                    run_code_line_number += 1
            else:
                for line in cc.code_lines:
                    run_code_list.append(f'{self.lang_def.chunk_wrapper_code_indent}{line}\n')
                    self.run_code_to_origins[run_code_line_number] = CodeLineOrigin(chunk=cc, line_number=user_code_line_number)
                    user_code_line_number += 1
                    run_code_line_number += 1
            last_cc = cc
        if self.code_chunks[-1].options['complete']:
            run_code_list.append(
                self.lang_def.chunk_wrapper_after_code.format(
                    stdout_end_delim=stdout_end_delim_template.format(chunk=last_cc.index, output_chunk=last_cc.output_index),
                    stderr_end_delim=stderr_end_delim_template.format(chunk=last_cc.index, output_chunk=last_cc.output_index),
                    repl_end_delim=repl_end_delim_template.format(chunk=last_cc.index, output_chunk=last_cc.output_index),
                )
            )
            self.expected_stdout_end_delim_chunks[last_cc.index] = 1
            self.expected_stderr_end_delim_chunks[last_cc.index] = 1
        if not self.code_chunks[-1].options['outside_main']:
            run_code_list.append(self.lang_def.run_template_after_code)
        self._run_code = ''.join(run_code_list)
        return self._run_code



# -*- coding: utf-8 -*-
#
# Copyright (c) 2021-2022, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the BSD 3-Clause License:
# http://opensource.org/licenses/BSD-3-Clause
#

'''
Python REPL emulation for Codebraid.  Reads code interspersed with delimiters
from stdin.  Writes code plus output interspersed with delimiters to stdout.
'''


import code
import io
import sys


# Prevent apport systems in some Linux distributions from breaking
# exception handling
sys.excepthook = sys.__excepthook__

# A sequence of ASCII characters at the start of any line that is a
# Codebraid delimiter line.
delim_start = sys.argv[1]
# The unique hash (hex) that identifies the code being executed.  This is
# present in any delimiter line and is used to avoid `delim_start` causing
# false positives in identifying delimiter lines.
hash = sys.argv[2]
# Optional buffering mode.
buffering = sys.argv[3] if len(sys.argv) > 3 else 'line'


sys.stdin = io.TextIOWrapper(sys.stdin.buffer, encoding='utf8')
if buffering == 'line':
    sys.stdout = io.TextIOWrapper(sys.stdout.buffer, encoding='utf8', line_buffering=True)
    sys.stderr = io.TextIOWrapper(sys.stderr.buffer, encoding='utf8', line_buffering=True)
elif buffering == 'unbuffered':
    sys.stdout = io.TextIOWrapper(sys.stdout.buffer, encoding='utf8', write_through=True)
    sys.stderr = io.TextIOWrapper(sys.stderr.buffer, encoding='utf8', write_through=True)
elif buffering == 'buffered':
    sys.stdout = io.TextIOWrapper(sys.stdout.buffer, encoding='utf8')
    sys.stderr = io.TextIOWrapper(sys.stderr.buffer, encoding='utf8')
else:
    raise ValueError

def parse_delim(delim_line: str):
    start = delim_line.find('(')
    end = delim_line.rfind(')')
    if start == -1 or end == -1:
        raise ValueError
    try:
        kv_pairs = [x.strip() for x in delim_line[start+1:end].split(',') if x]
        k_v_dict = {}
        for kv in kv_pairs:
            k, v = kv.split('=')
            k_v_dict[k] = v
        for k in ('chunk', 'output_chunk'):
            k_v_dict[k] = int(k_v_dict[k])
    except Exception:
        raise ValueError
    if k_v_dict['hash'] != hash:
        raise ValueError
    return k_v_dict




class Console(code.InteractiveConsole):
    def __init__(self):
        super().__init__(filename='<stdin>')
        self._line_iter = (line.rstrip('\n') for line in sys.stdin)
        self._cached_line = None
        self._last_line = None

    def raw_input(self, prompt):
        line = self._cached_line
        self._cached_line = None
        while True:
            if line is None:
                try:
                    line = next(self._line_iter)
                except StopIteration:
                    raise EOFError
            if line.startswith(delim_start):
                try:
                    delim_dict = parse_delim(line)
                except Exception:
                    break
                if prompt == sys.ps2:
                    if (delim_dict['delim'] == 'end' and delim_dict['output'] == 'repl' and
                            delim_dict['chunk'] == delim_dict['output_chunk'] and self._last_line):
                        self._cached_line = line
                        line = ''
                        break
                    raise EOFError
                if delim_dict['output'] == 'repl':
                    print(f'\n{line}')
                elif delim_dict['output'] == 'stdout':
                    print(f'\n{line}', flush=True)
                elif delim_dict['output'] == 'stderr':
                    print(f'\n{line}', file=sys.stderr, flush=True)
                else:
                    raise ValueError
                line = None
                continue
            break
        self._last_line = line
        if line:
            print(f'{prompt}{line}')
        else:
            print(prompt.rstrip())
        return line

    def write(self, text):
        print(text, end='')


console = Console()
console.interact(banner='', exitmsg='')



# -*- coding: utf-8 -*-
#
# Copyright (c) 2021-2022, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the BSD 3-Clause License:
# http://opensource.org/licenses/BSD-3-Clause
#


from __future__ import annotations


import asyncio
import collections
import io
import pathlib
import platform
import re
import shlex
import shutil
import subprocess
import tempfile
from typing import Optional
from .. import util
from .. import message
from ..code_collections import Session, CodeLineOrigin
from ..code_chunks import CodeChunk
from ..progress import Progress




async def exec(session: Session, *, cache_key_path: pathlib.Path, progress: Progress) -> None:
    '''
    Execute code from a session with the built-in code execution system,
    attach textual output to the code chunks within the session, and save rich
    output files.
    '''
    session.did_exec = True

    with tempfile.TemporaryDirectory() as temp_dir_str:
        temp_dir_path = pathlib.Path(temp_dir_str)
        if not session.lang_def.interpreter_script:
            origin_path = temp_dir_path / f'source_{session.hash_root}.{session.lang_def.extension}'
            origin_path.write_text(session.run_code, encoding='utf8')
        for stage, command_or_commands in session.lang_def.exec_stages.items():
            if session.status.prevent_exec:
                break
            progress.session_exec_stage_start(session, stage=stage)
            if isinstance(command_or_commands, str):
                commands = [command_or_commands]
            else:
                commands = command_or_commands
            for command in commands:
                if session.status.prevent_exec:
                    progress.session_exec_stage_end(session, stage=stage)
                    break
                subproc = SessionSubprocess(session, stage, command, temp_dir_path, progress)
                await subproc.start()
                await subproc.wait()
            progress.session_exec_stage_end(session, stage=stage)
        progress.session_finished(session)


class SessionSubprocess(object):
    def __init__(self, session: Session, stage: str, command: str, temp_dir_path: pathlib.Path, progress: Progress):
        self.session = session
        self.stage = stage
        self.encoding = getattr(session.lang_def, f'{stage}_encoding')
        self.has_interpreter_script = session.lang_def.interpreter_script is not None
        template_dict = {
            'executable': session.executable,
            'extension': session.lang_def.extension,
            'run_delim_start': session.run_delim_start,
            'run_delim_hash': session.run_delim_hash,
        }
        if not self.has_interpreter_script:
            origin_name = f'source_{session.hash_root}'
            origin_path = temp_dir_path / f'{origin_name}.{session.lang_def.extension}'
            template_dict.update({
                'source_name': origin_name,
                'source_dir': temp_dir_path.as_posix(),
                'source': origin_path.as_posix(),
                'source_without_extension': (temp_dir_path / origin_name).as_posix(),
            })
        else:
            template_dict.update({
                'run_script': session.lang_def.interpreter_script.as_posix(),
                'buffering': 'line',
            })
        program_with_args: list[str] = []
        for s in shlex.split(command):
            if s == '{executable_opts}':
                if session.executable_opts:
                    program_with_args.extend(session.executable_opts)
                continue
            if s == '{args}':
                if session.args:
                    program_with_args.extend(session.args)
                continue
            program_with_args.append(s.format(**template_dict))
        if platform.system() == 'Windows':
            # Modify args since subprocess.Popen() ignores PATH
            #   * https://bugs.python.org/issue8557
            #   * https://bugs.python.org/issue15451
            program = shutil.which(program_with_args[0]) or program_with_args[0]
        else:
            program = program_with_args[0]
        self.program = program
        self.args = program_with_args[1:]
        self.stage = stage
        self.progress = progress
        self.input = session.run_code if self.has_interpreter_script else None
        self.stderr_is_stdout = (stage != 'run')

        self._stdout_buffer: list[bytes] = []
        self._stderr_buffer: list[bytes] = []
        self._delim_error: bool = False
        self._sync_chunk_start_delims_state: dict[int, int] = collections.defaultdict(int)
        self._sync_chunk_end_delims_state: dict[int, int] = collections.defaultdict(int)
        self._sync_chunk_start_delims_waiting: int = 0
        self._sync_chunk_end_delims_waiting: int = 0
        self._sync_stream_end_waiting: int = 0

        if not self.has_interpreter_script:
            origin_path_re_pattern = re.escape(origin_path.as_posix()).replace('/', r'[\\/]')
            origin_name_re_pattern = re.escape(origin_path.name)
            self.origin_path_re = re.compile(f'{origin_path_re_pattern}|{origin_name_re_pattern}', re.IGNORECASE)
            self.origin_path_replacement = f'source.{session.lang_def.extension}'
            self.origin_path_inline_replacement = '<string>'
            self.line_number_pattern_re = session.lang_def.line_number_pattern_re
            self.line_number_regex_re = session.lang_def.line_number_regex_re
        self.error_patterns = session.lang_def.error_patterns
        self.warning_patterns = session.lang_def.warning_patterns
        home_path_re_pattern = re.escape(pathlib.Path('~').expanduser().as_posix()).replace('/', r'[\\/]')
        self.home_path_re = re.compile(home_path_re_pattern, re.IGNORECASE)

    @property
    def returncode(self):
        return self.proc.returncode

    @property
    def is_terminated(self):
        return self.proc.returncode is not None


    async def start(self):
        self.proc = None
        try:
            self.proc = await asyncio.create_subprocess_exec(
                self.program, *self.args,
                stdin=subprocess.PIPE if self.input else None,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT if self.stderr_is_stdout else subprocess.PIPE
            )
        except FileNotFoundError:
            self.session.errors.append(message.SysConfigError(f'Could not find executable "{self.program}"'))
            return
        if self.stage != 'run':
            await asyncio.gather(self._read_output())
        else:
            await asyncio.gather(self._write_input(), self._read_run('stdout'), self._read_run('stderr'))

    async def wait(self):
        if self.proc is None:
            return
        await self.proc.wait()
        for error in self.session.errors:
            if isinstance(error, message.ExecError):
                error.exit_code = self.proc.returncode
        for cc in self.session.code_chunks:
            for error in cc.errors:
                if isinstance(error, message.ExecError):
                    error.exit_code = self.proc.returncode
        return self.proc.returncode


    def _decode(self, output: bytes, *, output_type: Optional[str]=None, code_chunk: Optional[CodeChunk]=None):
        # Decode bytes using `io` to get the same newline behavior as reading
        # from a file.  Only raise decoding errors for the run stage, since
        # that is the only stage that produces output in the document.  At
        # some point, it may be worth adding a warning for decoding errors for
        # other stages.
        if self.stage != 'run':
            return io.TextIOWrapper(io.BytesIO(output), encoding=self.encoding, errors='backslashreplace').read()
        try:
            decoded = io.TextIOWrapper(io.BytesIO(output), encoding=self.encoding).read()
        except UnicodeDecodeError as e:
            self.session.decode_error_count += 1
            if self.session.decode_error_count <= self.session.max_tracked_decode_error_count:
                error = message.DecodeError(
                    f'Error decoding {output_type or "output"} as "{self.encoding}" (invalid bytes shown in \\xNN format):\n{e}'
                )
                if code_chunk is not None:
                    code_chunk.errors.append(error)
                else:
                    self.session.errors.append(error)
            decoded = io.TextIOWrapper(io.BytesIO(output), encoding=self.encoding, errors='backslashreplace').read()
        return decoded


    async def _write_input(self):
        if not self.input:
            return
        for line in io.BytesIO(self.input.encode(self.encoding)):
            self.proc.stdin.write(line)
            await self.proc.stdin.drain()
        self.proc.stdin.close()
        await self.proc.stdin.wait_closed()


    async def _read_output(self):
        while True:
            out = await self.proc.stdout.read(4096)
            if out:
                self._stdout_buffer.append(out)
            else:
                break
        stdout = self._decode(b''.join(self._stdout_buffer))
        if self.stage == 'compile':
            stdout = self._sync_stderr_or_compile_output(
                stdout, code_chunk=None, session_output_lines=self.session.compile_lines
            )
            self.session.compile_lines.extend(util.splitlines_lf(stdout))
        elif self.stage == 'pre_run':
            self.session.pre_run_output_lines.extend(util.splitlines_lf(stdout))
            self.session.errors.append(message.PreRunError(self.session.pre_run_output_lines))
        elif self.stage == 'post_run':
            self.session.post_run_output_lines.extend(util.splitlines_lf(stdout))
            self.session.errors.append(message.PostRunError(self.session.post_run_output_lines))
        else:
            raise ValueError
        self.progress.session_exec_stage_output(self.session, output=stdout)

    def _parse_delim(self, delim_line: str):
        start = delim_line.find('(')
        end = delim_line.rfind(')')
        if start == -1 or end == -1:
            raise ValueError
        try:
            kv_pairs = [x.strip() for x in delim_line[start+1:end].split(',') if x]
            k_v_dict = {}
            for kv in kv_pairs:
                k, v = kv.split('=')
                k_v_dict[k] = v
            for k in ('chunk', 'output_chunk'):
                k_v_dict[k] = int(k_v_dict[k])
        except Exception:
            raise ValueError
        if k_v_dict['hash'] != self.session.run_delim_hash:
            raise ValueError
        return k_v_dict

    def _process_code_chunk_output(self, output: bytes, *, code_chunk: CodeChunk, output_type: str):
        if not output:
            return
        output_str = self._decode(output, output_type=output_type, code_chunk=code_chunk)
        if output_type == 'stderr':
            output_str = self._sync_stderr_or_compile_output(output_str, code_chunk=code_chunk)
        getattr(self.progress, f'session_chunk_{output_type}')(self.session, chunk=code_chunk, output=output_str)
        getattr(code_chunk, f'{output_type}_lines').extend(util.splitlines_lf(output_str))

    def _run_line_number_to_origin(self, run_line_number: int) -> CodeLineOrigin | tuple[None, None]:
        line_origin = self.session.run_code_to_origins.get(run_line_number, None)
        while line_origin is None and run_line_number > 0:
            run_line_number -= 1
            origin = self.session.run_code_to_origins.get(run_line_number, None)
        if line_origin is None:
            return (None, None)
        return line_origin

    def _sync_stderr_or_compile_output(self, output: str, *,
                                       code_chunk: Optional[CodeChunk],
                                       session_output_lines: Optional[list[str]]=None) -> str:
        if not ((code_chunk is not None and session_output_lines is None) or
                (code_chunk is None and session_output_lines is not None)):
            raise TypeError

        if (self.has_interpreter_script or self.origin_path_re.search(output) is None and
                (self.line_number_regex_re is None or self.line_number_regex_re.search(output) is None)):
            output = self.home_path_re.sub('~', output)
            for error_pattern in self.error_patterns:
                if error_pattern in output:
                    if code_chunk is not None:
                        if not code_chunk.errors.has_stderr:
                            code_chunk.errors.append(message.StderrRunError(code_chunk.stderr_lines))
                    elif not self.session.errors.has_ref(session_output_lines):
                        self.session.errors.append(message.StderrRunError(session_output_lines))
                    return output
            for warning_pattern in self.warning_patterns:
                if warning_pattern in output:
                    if code_chunk is not None:
                        if not code_chunk.warnings.has_stderr:
                            code_chunk.warnings.append(message.StderrRunWarning(code_chunk.stderr_lines))
                    elif not self.session.warnings.has_ref(session_output_lines):
                        self.session.warnings.append(message.StderrRunWarning(session_output_lines))
                    return output
            return output

        # For each line containing the origin (source) path, replace the path
        # with a generic file name.  Sync line numbers with the line numbers
        # from user code (existing line numbers include template code).  In
        # the event of line numbers that can't be synchronized, wrap numbers
        # in square brackets `[<number>]`.
        max_synced_code_chunk = None
        if not self.line_number_pattern_re:
            if code_chunk is not None and code_chunk.inline:
                output = self.origin_path_re.sub(self.origin_path_inline_replacement, output)
            else:
                output = self.origin_path_re.sub(self.origin_path_replacement, output)
        else:
            output_lines_indices_with_origin_path = set()
            def process_match(match: re.Match):
                line_number = output.count('\n', 0, match.start())
                output_lines_indices_with_origin_path.add(line_number)
                return self.origin_path_replacement
            output = self.origin_path_re.sub(process_match, output)
            output_lines = util.splitlines_lf(output)
            if output.endswith('\n'):
                output_lines.append('')
            for index in output_lines_indices_with_origin_path:
                line = output_lines[index]
                current_code_chunk = None
                line_list = []
                last_end = 0
                for line_number_match in self.line_number_pattern_re.finditer(line):
                    for group_index in range(1, line_number_match.lastindex+1):
                        if line_number_match.group(group_index) is not None:
                            line_list.append(line[last_end:line_number_match.start(group_index)])
                            synced_code_chunk, synced_line_number = self._run_line_number_to_origin(int(line_number_match.group(group_index)))
                            if synced_code_chunk is None:
                                current_code_chunk = None
                                line_list.append(f'[{line_number_match.group(group_index)}]')
                            else:
                                current_code_chunk = synced_code_chunk
                                if max_synced_code_chunk is None or max_synced_code_chunk.index > synced_code_chunk.index:
                                    max_synced_code_chunk = synced_code_chunk
                                line_list.append(f'{synced_line_number}')
                            line_list.append(line[line_number_match.end(group_index):line_number_match.end()])
                            last_end = line_number_match.end()
                            break
                line_list.append(line[last_end:])
                line = ''.join(line_list)
                if current_code_chunk is not None and current_code_chunk.inline:
                    # Inline replacements must be performed last, so that
                    # patterns like `.ext:{number}` will work.  Otherwise,
                    # this has become `<string>:{number}` and matching fails.
                    line = line.replace(self.origin_path_replacement, self.origin_path_inline_replacement)
                output_lines[index] = line
            output = '\n'.join(output_lines)
        if self.line_number_regex_re:
            output_list = []
            last_end = 0
            for line_number_match in self.line_number_regex_re.finditer(output):
                for group_index in range(1, line_number_match.lastindex+1):
                    if line_number_match.group(group_index) is not None:
                        output_list.append(output[last_end:line_number_match.start(group_index)])
                        _, synced_line_number = self._run_line_number_to_origin(int(line_number_match.group(group_index)))
                        if synced_line_number is None:
                            output_list.append(f'[{line_number_match.group(group_index)}]')
                        else:
                            output_list.append(f'{synced_line_number:{line_number_match.end(group_index)-line_number_match.start(group_index)}d}')
                        output_list.append(output[line_number_match.end(group_index):line_number_match.end()])
                        last_end = line_number_match.end()
                        break
            output_list.append(output[last_end:])
            output = ''.join(output_list)
        # Wait to sanitize home dir until after normalizing temp paths
        output = self.home_path_re.sub('~', output)
        for error_pattern in self.error_patterns:
            if error_pattern in output:
                if code_chunk is not None:
                    if not code_chunk.errors.has_stderr:
                        code_chunk.errors.append(message.StderrRunError(code_chunk.stderr_lines))
                else:
                    if not self.session.errors.has_ref(session_output_lines):
                        self.session.errors.append(message.StderrRunError(session_output_lines))
                    if max_synced_code_chunk is not None and not max_synced_code_chunk.errors.has_ref(session_output_lines):
                        max_synced_code_chunk.errors.append(message.StderrRunErrorRef(session_output_lines))
                        self.session.errors.update_refed(session_output_lines)
                return output
        for warning_pattern in self.warning_patterns:
            if warning_pattern in output:
                if code_chunk is not None:
                    if not code_chunk.warnings.has_stderr:
                        code_chunk.warnings.append(message.StderrRunWarning(code_chunk.stderr_lines))
                else:
                    if not self.session.warnings.has_ref(session_output_lines):
                        self.session.warnings.append(message.StderrRunWarning(session_output_lines))
                    if max_synced_code_chunk is not None and not max_synced_code_chunk.warnings.has_ref(session_output_lines):
                        max_synced_code_chunk.warnings.append(message.StderrRunWarningRef(session_output_lines))
                return output
        return output


    async def _sync_chunk_start_delims(self, code_chunk: CodeChunk):
        self._sync_chunk_start_delims_state[code_chunk.index] += 1
        if self._sync_chunk_start_delims_state[code_chunk.index] == 2:
            self.progress.session_chunk_start_exec(self.session, chunk=code_chunk)
            return
        self._sync_chunk_start_delims_waiting += 1
        while self._sync_chunk_start_delims_state[code_chunk.index] < 2:
            if self._delim_error:
                break
            if self._sync_chunk_start_delims_waiting > 1:
                code_chunk.errors.append(message.RuntimeSourceError(
                    'Synchronization of code output with document failed.  Possible modification of stdout or stderr?'
                ))
                self._delim_error = True
                break
            await asyncio.sleep(0)
        self._sync_chunk_start_delims_waiting -= 1

    async def _sync_chunk_end_delims(self, code_chunk: CodeChunk):
        self._sync_chunk_end_delims_state[code_chunk.index] += 1
        if self._sync_chunk_end_delims_state[code_chunk.index] == 2:
            self.progress.session_chunk_end_exec(self.session, chunk=code_chunk)
            return
        self._sync_chunk_end_delims_waiting += 1
        while self._sync_chunk_end_delims_state[code_chunk.index] < 2:
            if self._delim_error:
                break
            if self._sync_chunk_end_delims_waiting > 1:
                code_chunk.errors.append(message.RuntimeSourceError(
                    'Synchronization of code output with document failed.  Possible modification of stdout or stderr?'
                ))
                self._delim_error = True
                break
            await asyncio.sleep(0)
        self._sync_chunk_end_delims_waiting -= 1

    async def _sync_stream_end(self, stream_type: str):
        # Add sleep for stdout to make order of stream resolution reproducible
        self._sync_stream_end_waiting += 1
        if self._sync_stream_end_waiting == 2:
            if stream_type == 'stdout':
                await asyncio.sleep(0)
            return
        while self._sync_stream_end_waiting < 2:
            await asyncio.sleep(0)
        self._sync_stream_end_waiting = 0
        if stream_type == 'stdout':
            await asyncio.sleep(0)


    async def _read_run(self, stream_type: str):
        # Currently, progress is line buffered.

        session = self.session
        encoding = self.encoding
        run_delim_start_search_pattern = self.session.run_delim_start_search_pattern.encode(encoding)
        lf = '\n'.encode(encoding)
        lflf = 2 * lf
        cr = '\r'.encode(encoding)
        crlf = cr + lf
        crflcrlf = 2 * crlf
        # Assume encoding with len(cr) == len(lf)
        len_cr_or_lf = len(lf)
        if stream_type == 'stdout':
            stream = self.proc.stdout
            buffer = self._stdout_buffer
        elif stream_type == 'stderr':
            stream = self.proc.stderr
            buffer = self._stderr_buffer
        else:
            raise ValueError

        if session.code_chunks[0].options['outside_main']:
            first_code_chunk = session.code_chunks[0]
            code_chunk = session.code_chunks[first_code_chunk.output_index]
            await self._sync_chunk_start_delims(first_code_chunk)
        else:
            code_chunk = None
        output_type = stream_type

        while True:
            output = await stream.read(4096)
            if self._delim_error:
                if output:
                    continue
                break
            if not output:
                if buffer:
                    remaining_output = b''.join(buffer)
                    buffer.clear()
                    if code_chunk is not None:
                        if code_chunk.options['outside_main'] and code_chunk is session.code_chunks[-1]:
                            self._process_code_chunk_output(remaining_output, code_chunk=code_chunk, output_type=output_type)
                            await self._sync_chunk_end_delims(code_chunk)
                        else:
                            self._process_code_chunk_output(remaining_output, code_chunk=code_chunk, output_type=output_type)
                            await self._sync_chunk_end_delims(code_chunk)
                            self._delim_error = True
                            if not code_chunk.errors.has_stderr:
                                code_chunk.errors.append(message.RuntimeSourceError(
                                    'Code chunk is not a complete unit of code or exited before expected.'
                                ))
                    elif stream_type == 'stderr':
                        if (all(v == 0 for v in session.expected_stderr_start_delim_chunks.values()) and
                                all(v == 0 for v in session.expected_stderr_end_delim_chunks.values())):
                            session_lines = session.template_end_stderr_lines
                        else:
                            session_lines = session.other_stderr_lines
                        synced_output = self._sync_stderr_or_compile_output(
                            self._decode(remaining_output, output_type='stderr', code_chunk=None),
                            code_chunk=None,
                            session_output_lines=session_lines
                        )
                        session_lines.extend(util.splitlines_lf(synced_output))
                elif code_chunk is not None:
                    await self._sync_chunk_end_delims(code_chunk)
                await self._sync_stream_end(stream_type)

                # Only report a single error as a result of missing
                # delimiter(s), for the first code chunk with a missing
                # delimiter.
                if not self._delim_error and not session.status.prevent_exec:
                    for code_chunk_index, count in getattr(session, f'expected_{stream_type}_start_delim_chunks').items():
                        if count != 0:
                            session.code_chunks[code_chunk_index].errors.append(message.RuntimeSourceError(
                                "A previous code chunk interfered with this chunk's execution.  "
                                'Incorrect "outside_main" or "complete" settings, or incomplete preceding unit of code.'
                            ))
                            break
                if not self._delim_error and not session.status.prevent_exec:
                    for code_chunk_index, count in getattr(session, f'expected_{stream_type}_end_delim_chunks').items():
                        if count != 0:
                            session.code_chunks[code_chunk_index].errors.append(message.RuntimeSourceError(
                                'Code chunk is not a complete unit of code or exited before expected.'
                            ))
                            break
                break
            buffer.append(output)
            if lf not in output and cr not in output:
                continue
            unprocessed_output = b''.join(buffer)
            buffer.clear()
            delim_start_index = unprocessed_output.find(run_delim_start_search_pattern)
            if delim_start_index == -1:
                if code_chunk is not None:
                    # Leave a trailing `\r` since it might be followed by
                    # `\n`.  Leave a trailing `\n` since it might be from
                    # delim rather that user.
                    last_lf_index = unprocessed_output.rfind(lf)
                    last_cr_index = unprocessed_output.rfind(cr, 0, len(unprocessed_output) - len_cr_or_lf)
                    break_index = max(last_lf_index, last_cr_index)
                    if break_index != -1:
                        output_before_break = unprocessed_output[:break_index+len_cr_or_lf]
                        unprocessed_output = unprocessed_output[break_index+len_cr_or_lf:]
                        if output_before_break:
                            self._process_code_chunk_output(output_before_break, code_chunk=code_chunk, output_type=output_type)
            else:
                while delim_start_index != -1:
                    output_before_delim = unprocessed_output[:delim_start_index]
                    if output_before_delim == lf or output_before_delim == crlf:
                        output_before_delim = b''
                    elif output_before_delim.endswith(lflf):
                        output_before_delim = output_before_delim[:-len_cr_or_lf]
                    elif output_before_delim.endswith(crflcrlf):
                        output_before_delim = output_before_delim[:-2*len_cr_or_lf]
                    if output_before_delim:
                        if code_chunk is not None:
                            self._process_code_chunk_output(output_before_delim, code_chunk=code_chunk, output_type=output_type)
                    delim_end_index = unprocessed_output.find(lf, delim_start_index)
                    if delim_end_index == -1:
                        break
                    delim = unprocessed_output[delim_start_index:delim_end_index+len_cr_or_lf]
                    unprocessed_output = unprocessed_output[delim_end_index+len_cr_or_lf:]
                    try:
                        delim_dict = self._parse_delim(delim.decode(encoding))
                    except Exception:
                        if code_chunk is not None:
                            self._process_code_chunk_output(delim, code_chunk=code_chunk, output_type=output_type)
                        delim_start_index = -1
                    else:
                        if delim_dict['delim'] == 'start':
                            delim_output_type = delim_dict['output']
                            if delim_output_type == stream_type:
                                expected_start_delim_chunks = getattr(session, f'expected_{stream_type}_start_delim_chunks')
                                expected_start_delim_chunks[delim_dict['chunk']] -= 1
                                if output_type != stream_type or code_chunk is not None or expected_start_delim_chunks[delim_dict['chunk']] != 0:
                                    self._delim_error = True
                                    session.code_chunks[delim_dict['chunk']].errors.append(message.RuntimeSourceError(
                                        "A previous code chunk interfered with this chunk's execution.  "
                                        'Incorrect "outside_main" or "complete" settings, or incomplete preceding unit of code.'
                                    ))
                                    break
                                code_chunk = session.code_chunks[delim_dict['output_chunk']]
                                await self._sync_chunk_start_delims(session.code_chunks[delim_dict['chunk']])
                            else:
                                if output_type != stream_type or code_chunk is None:
                                    self._delim_error = True
                                    session.code_chunks[delim_dict['chunk']].errors.append(message.RuntimeSourceError(
                                        'Code chunk is not a complete unit of code or is invalid.'
                                    ))
                                    break
                                output_type = delim_output_type
                        elif delim_dict['delim'] == 'end':
                            delim_output_type = delim_dict['output']
                            if delim_output_type == stream_type:
                                expected_end_delim_chunks = getattr(session, f'expected_{stream_type}_end_delim_chunks')
                                expected_end_delim_chunks[delim_dict['chunk']] -= 1
                                if output_type != stream_type or code_chunk is None or code_chunk.output_index != delim_dict['output_chunk'] or expected_end_delim_chunks[delim_dict['chunk']] != 0:
                                    self._delim_error = True
                                    session.code_chunks[delim_dict['chunk']].errors.append(message.RuntimeSourceError(
                                        'Code chunk is not a complete unit of code or is invalid.'
                                    ))
                                    break
                                code_chunk = None
                                await self._sync_chunk_end_delims(session.code_chunks[delim_dict['chunk']])
                            else:
                                if output_type != delim_output_type or code_chunk is None:
                                    self._delim_error = True
                                    session.code_chunks[delim_dict['chunk']].errors.append(message.RuntimeSourceError(
                                        'Code chunk is not a complete unit of code or is invalid.'
                                    ))
                                    break
                                output_type = stream_type
                        else:
                            raise ValueError
                        delim_start_index = unprocessed_output.find(run_delim_start_search_pattern)
            if unprocessed_output:
                buffer.append(unprocessed_output)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018-2022, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the BSD 3-Clause License:
# http://opensource.org/licenses/BSD-3-Clause
#


from __future__ import annotations


import collections
import collections.abc
import hashlib
import io
import json
import os
import pathlib
from typing import Optional, Sequence, Union
import zipfile
from .. import codeprocessors
from ..codebraid_defaults import CodebraidDefaults
from ..progress import Progress




class MetaConverter(type):
    '''
    Metaclass for converters.  Allows converters to register themselves
    by name and by compatible formats.
    '''
    def __init__(cls, name, bases, dct):
        if not hasattr(cls, '_registry'):
            # Base Converter class
            cls._registry = {}
        else:
            # Subclass
            cls._registry[name.lower()] = cls
            if not all(attr is None or
                       (isinstance(attr, set) and attr and all(isinstance(x, str) for x in attr))
                       for attr in [cls.from_formats, cls.multi_origin_formats, cls.to_formats]):
                raise TypeError
            if (cls.from_formats is not None and cls.multi_origin_formats is not None and
                    cls.multi_origin_formats - cls.from_formats):
                raise ValueError
        super().__init__(name, bases, dct)




class Converter(object):
    '''
    Base class for converters.
    '''
    __metaclass__ = MetaConverter

    def __init__(self, *,
                 strings: Optional[Union[str, Sequence[str]]]=None,
                 string_origins: Optional[Union[str, pathlib.Path, Sequence[str], Sequence[pathlib.Path]]]=None,
                 paths: Optional[Union[str, Sequence[str], pathlib.Path, Sequence[pathlib.Path]]]=None,
                 no_cache: Optional[bool]=False,
                 cache_path: Optional[Union[str, pathlib.Path]]=None,
                 cross_origin_sessions: bool=True,
                 expanduser: bool=False,
                 expandvars: bool=False,
                 from_format: Optional[str]=None,
                 codebraid_defaults: Optional[CodebraidDefaults]=None,
                 no_execute: bool=False,
                 only_code_output: Optional[str]=None,
                 synctex: bool=False):

        if not all(isinstance(x, bool) for x in (cross_origin_sessions, expanduser, expandvars)):
            raise TypeError
        self.cross_origin_sessions = cross_origin_sessions
        self.expanduser = expanduser
        self.expandvars = expandvars
        if codebraid_defaults is None:
            codebraid_defaults = CodebraidDefaults()
        elif not isinstance(codebraid_defaults, CodebraidDefaults):
            raise TypeError
        self.codebraid_defaults = codebraid_defaults

        if paths is not None and strings is None:
            if string_origins is not None:
                raise TypeError
            if isinstance(paths, str):
                paths = [pathlib.Path(paths)]
            elif isinstance(paths, pathlib.Path):
                paths = [paths]
            elif isinstance(paths, collections.abc.Sequence) and paths:
                if all(isinstance(x, str) for x in paths):
                    paths = [pathlib.Path(x) for x in paths]
                elif not all(isinstance(x, pathlib.Path) for x in paths):
                    raise TypeError
            else:
                raise TypeError
            self.raw_origin_paths = paths
            # Names are based on paths BEFORE any expansion
            origin_names = [p.as_posix() for p in paths]
            if not all(isinstance(x, bool) for x in (expanduser, expandvars)):
                raise TypeError
            if expandvars:
                paths = [pathlib.Path(os.path.expandvars(str(p))) for p in paths]
            if expanduser:
                paths = [p.expanduser() for p in paths]
            self.expanded_origin_paths = collections.OrderedDict(zip(origin_names, paths))
            origin_strings = []
            for p in paths:
                try:
                    origin_string = p.read_text(encoding='utf_8_sig')
                except Exception as e:
                    if not p.is_file():
                        raise ValueError('File "{0}" does not exist'.format(p))
                    raise ValueError('Failed to read file "{0}":\n  {1}'.format(p, e))
                if not origin_string:
                    origin_string = '\n'
                origin_strings.append(origin_string)
            self.origins = collections.OrderedDict(zip(origin_names, origin_strings))
            if self.from_formats is not None:
                if from_format is None:
                    try:
                        origin_formats = set([self._file_extension_to_format_dict[p.suffix] for p in paths])
                    except KeyError:
                        raise TypeError('Cannot determine document format from file extensions, or unsupported format')
                    from_format = origin_formats.pop()
                    if origin_formats:
                        raise TypeError('Cannot determine unambiguous document format from file extensions')
                if from_format not in self.from_formats:
                    raise ValueError('Unsupported document format {0}'.format(from_format))
            self.from_format = from_format
        elif strings is not None and paths is None:
            if not all(x is False for x in (expanduser, expandvars)):
                if not all(isinstance(x, bool) for x in (expanduser, expandvars)):
                    raise TypeError
                raise ValueError
            if isinstance(strings, str):
                strings = [strings]
            elif not (isinstance(strings, collections.abc.Sequence) and
                      strings and all(isinstance(x, str) for x in strings)):
                raise TypeError
            string_origins_normalized: Optional[Union[Sequence[str], Sequence[pathlib.Path]]]
            if string_origins is None:
                string_origins_normalized = None
            elif isinstance(string_origins, str) or isinstance(string_origins, pathlib.Path):
                string_origins_normalized = [string_origins]
            elif (isinstance(string_origins, collections.abc.Sequence) and string_origins and
                    (all(isinstance(x, str) for x in string_origins) or all(isinstance(x, pathlib.Path) for x in string_origins))):
                string_origins_normalized = string_origins
            else:
                raise TypeError

            # Normalize newlines, as if read from file with universal newlines
            origin_strings = []
            for s in strings:
                if '\r' in s:
                    origin_strings.append(io.StringIO(s, newline=None).read() or '\n')
                else:
                    origin_strings.append(s or '\n')
            if string_origins_normalized is not None:
                origin_names = []
                self.raw_origin_paths = []
                self.expanded_origin_paths = {}
                cwd_path = pathlib.Path('.').absolute()
                for string_origin in string_origins_normalized:
                    string_origin_path = pathlib.Path(string_origin)
                    try:
                        string_origin_rel_path = string_origin_path.relative_to(cwd_path)
                    except ValueError:
                        string_origin_rel_path = string_origin_path
                    origin_name = string_origin_rel_path.as_posix()
                    origin_names.append(origin_name)
                    self.raw_origin_paths.append(string_origin_path)
                    self.expanded_origin_paths[origin_name] = string_origin_path
            else:
                if len(strings) == 1:
                    origin_names = ['<string>']
                else:
                    origin_names = ['<string({0})>'.format(n+1) for n in range(len(strings))]
                self.raw_origin_paths = None
                self.expanded_origin_paths = None
            self.origins = collections.OrderedDict(zip(origin_names, origin_strings))
            if from_format is None:
                raise TypeError('Document format is required')
            if self.from_formats is not None and from_format not in self.from_formats:
                raise ValueError('Unsupported document format {0}'.format(from_format))
            self.from_format = from_format
        else:
            raise TypeError
        if len(self.origins) > 1 and from_format not in self.multi_origin_formats:
            raise TypeError('Multiple source files are not supported for format {0}'.format(from_format))

        if not isinstance(no_cache, bool):
            raise TypeError
        self.no_cache = no_cache
        if cache_path is None:
            cache_path = pathlib.Path('_codebraid')
        elif isinstance(cache_path, str):
            cache_path = pathlib.Path(cache_path)
        elif not isinstance(cache_path, pathlib.Path):
            raise TypeError
        if expandvars:
            cache_path = pathlib.Path(os.path.expandvars(cache_path.as_posix()))
        if expanduser:
            cache_path = cache_path.expanduser()
        self.cache_path = cache_path
        cache_key_hasher = hashlib.blake2b()
        if self.expanded_origin_paths is None:
            origin_paths_for_cache = None
            cache_key_hasher.update(b'<string>')
        else:
            origin_paths_for_cache = []
            for p in self.expanded_origin_paths.values():
                try:
                    p_final = pathlib.Path('~') / p.absolute().relative_to(pathlib.Path.home())
                except ValueError:
                    p_final = p.absolute()
                origin_paths_for_cache.append(p_final)
                cache_key_hasher.update(p_final.as_posix().encode('utf8'))
                cache_key_hasher.update(cache_key_hasher.digest())
        self.origin_paths_for_cache = origin_paths_for_cache
        self.cache_key = cache_key_hasher.hexdigest()[:16]

        if not isinstance(no_execute, bool):
            raise TypeError
        self.no_execute = no_execute

        if only_code_output is None:
            pass
        elif not isinstance(only_code_output, str):
            raise TypeError
        elif only_code_output not in ('codebraid_preview',):
            raise ValueError
        self.only_code_output = only_code_output

        self._io_map = False
        if not isinstance(synctex, bool):
            raise TypeError
        self.synctex = synctex
        if synctex:
            self._io_map = True

        self._progress = Progress(only_code_output=self.only_code_output, no_execute=self.no_execute)

        self.code_chunks = []
        self.code_processor: Optional[codeprocessors.CodeProcessor] = None


    def __enter__(self):
        self.code_braid()
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.cleanup()


    @property
    def exit_code(self):
        if self.code_processor is not None:
            return self.code_processor.exit_code
        return 0


    from_formats = None
    to_formats = None
    multi_origin_formats = None

    _file_extension_to_format_dict = {'.md': 'markdown', '.markdown': 'markdown',
                                      '.tex': 'latex', '.ltx': 'latex'}

    def code_braid(self):
        self.extract_code_chunks()
        self.process_code_chunks()
        if self.no_execute:
            pass
        else:
            self.exec_code_chunks()

    def extract_code_chunks(self):
        self._progress.parse_start()
        self._extract_code_chunks()
        self._progress.parse_end()

    def _extract_code_chunks(self):
        raise NotImplementedError

    def process_code_chunks(self):
        self._progress.process_start()
        self._process_code_chunks()
        self._progress.process_end()

    def _process_code_chunks(self):
        self.code_processor = codeprocessors.CodeProcessor(
            code_chunks=self.code_chunks,
            cross_origin_sessions=self.cross_origin_sessions,
            no_cache=self.no_cache,
            cache_path=self.cache_path,
            cache_key=self.cache_key,
            origin_paths_for_cache=self.origin_paths_for_cache,
            codebraid_defaults=self.codebraid_defaults,
            progress=self._progress,
            only_code_output=self.only_code_output,
        )
        self.code_processor.process()

    def exec_code_chunks(self):
        self._progress.exec_start()
        self.code_processor.exec()
        self._progress.exec_end()

    def convert(self, *, to_format, **kwargs):
        self._progress.convert_start()
        self._convert(to_format=to_format, **kwargs)
        self._progress.convert_end()

    def _convert(self, *, to_format):
        raise NotImplementedError

    def cleanup(self):
        if self.code_processor is not None:
            self.code_processor.cleanup()
        self._progress.complete()

    def _save_synctex_data(self, data):
        zip_path = self.cache_path / 'synctex.zip'
        with zipfile.ZipFile(str(zip_path), 'w', compression=zipfile.ZIP_DEFLATED) as zf:
            zf.writestr('synctex.json', json.dumps(data))



# -*- coding: utf-8 -*-
#
# Copyright (c) 2019-2022, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the BSD 3-Clause License:
# http://opensource.org/licenses/BSD-3-Clause
#


from __future__ import annotations


import collections
import random
from typing import Any, Callable




class KeyDefaultDict(collections.defaultdict):
    '''
    Default dict that passes missing keys to the factory function, rather than
    calling the factory function with no arguments.
    '''
    default_factory: Callable[[Any], Any]
    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        else:
            self[key] = self.default_factory(key)
            return self[key]


def random_ascii_lower_alpha(n):
    '''
    Create a random string of length n consisting of lowercase ASCII
    letters.  Useful for creating more robust tempfile names when working in
    `tempfile.TemporaryDirectory()`.
    '''
    return ''.join(chr(num) for num in (random.randrange(97, 122+1) for _ in range(n)))


def splitlines_lf(string, *, last_lf_stripped=False):
    r'''
    Like `str.splitlines()`, but only splits on `\n`.  This should be used
    on strings that have had `\r\n?` normalized to `\n`.  It avoids the
    `str.splitlines()` behavior of splitting on additional code points
    like `\v` and `\f`.
    '''
    lines = string.split('\n')
    if not last_lf_stripped and (string == '' or string[-1] == '\n'):
        lines.pop()
    return lines
